import java.util.*;
import soot.*;
import soot.jimple.*;

// TODO Check all the TODO entries in this file!
// One way to log information is to call a static helper method that you provide in a helper class
// (I called mine xlog.XLog).  Naturally, you will have to provide the helper jar file to Soot when
// instrumenting and to Java when running the benchmark.

public class Stub4 {
	public static final String[] PRELOAD_SIGNATURES = new String[] {
			"java.lang.LinkageError",
			"junit.framework.Assert",
			"java.util.AbstractQueue",
			"java.lang.ReflectiveOperationException",
			"java.util.AbstractCollection",
			"javax.imageio.IIOParam",
			"javax.imageio.spi.ImageReaderWriterSpi",
			// TODO: add the name of your helper class below
			// "xlog.XLog"
	};
	
	static SootMethod xlogger = null; // SootMethod that handles logging

	public static void main(String[] args) {
		Scene.v().setSootClassPath(null); // ensures that we load the class path from the command line
		Scene.v().addBasicClass("java.lang.LinkageError", SootClass.SIGNATURES); // Tell Soot to resolve this class in detail

		// Install your analysis, which is now executed (by default) on everything passed in the Soot classpath.
		PackManager.v().getPack("jtp").add(new Transform("jtp.smtransform", new CallgraphTransform()));
		for (String sig : PRELOAD_SIGNATURES) {
			Scene.v().addBasicClass(sig, SootClass.SIGNATURES);
		}

		soot.Main.main(args);
	}

	/**
	 * Assigns globally unique names to units
	 */
	public static final class UnitNamer {
		private String body_name;
		private HashMap<Unit, Integer> unit_numbers = new HashMap<>();
		
		/**
		 * Creates a UnitNamer for a given method body
		 * @param body
		 */
		public UnitNamer(Body body) {
			this.body_name = body.getMethod().getDeclaration();
			int count = 0;
			for (Unit u : body.getUnits()) {
				this.unit_numbers.put(u, count++);
			}
		}
		/**
		 * Looks up the name for the given unit
		 * @param u The unit to look up
		 * @return A suitable string name for the unit
		 */
		String getName(Unit u) {
			if (this.unit_numbers.containsKey(u)) {
				return this.body_name + "/" + this.unit_numbers.get(u);
			}
			throw new IllegalArgumentException();
		}
	}
	
	public static class CallgraphTransform extends BodyTransformer {
		
		@Override
		protected void internalTransform(Body body, String phaseName, Map<String, String> options) {
			System.out.println("Getting class:" + Scene.v().getSootClass("xlog.Xlog"));
			for (SootMethod m : Scene.v().getSootClass("xlog.Xlog").getMethods()) {
				System.out.println(m);
			}
			if (xlogger == null) {
				// TODO: Loads the SootMethod for logging, on demand.
				// If this fails, you may have forgotten to add the name of your helper class to PRELOAD_SIGNATURES 
				xlogger = Scene.v().getMethod("<xlog.XLog: void m(java.lang.Object,java.lang.String,java.lang.String)>");
			}
			// Assigns the same names to units that the static callgraph uses
			UnitNamer namer = new UnitNamer(body);

			// Mutate the method body
			PatchingChain<Unit> units = body.getUnits();
			for (UnitBox ub : body.getAllUnitBoxes()) {
				Unit u = ub.getUnit();
				// Produces the unit's name; this is equal to the name used in the static callgraph.
				String unit_name = namer.getName(u);
				
				// TODO: check if this unit is a method call.
				// If so, add suitable instrumentation. 
				// Example: this is how you can create a StaticInvokeExpr:
				//     StaticInvokeExpr instrumentation_invocation = Jimple.v().newStaticInvokeExpr(xlogger.makeRef(), receiver, StringConstant.v(a1), StringConstant.v(a2));
				// Example: this is how you can insert an additional statement into the method:
				//     units.insertBefore(Jimple.v().newInvokeStmt(instrumentation_invocation), u);
				//
				// Log the dynamic call graph.  Make sure that the information that you log matches the one in the static call graph (i.e.: unit_name + "\t" + target_callsite)
				// Here, target_callsite requires a bit of further work:
				//
				//   You will get the receiver object (cf. InstanceInvokeExpr.getBase()) at run-time.  Your helper method referenced
				// by xlogger can now use Java reflection (obj.getClass()) to figure out what the dynamic type of this object is.  Unfortunately this
				// part is a bit fiddly, since (a) Soot's method signatures look subtly different from Java's signatures, meaning that you can't look them
				// up directly, and (b) Java makes you work a little to find out what class it was that _declared_ a certain method.  Make sure that for the following
				// classes, your method will figure out that `dynamic class C' means `C.foo()', and `dynamic class B' means `A.foo()':
				//
				// class A {
				//   int foo() { print(1); }
				// }
				// class B extends A {} // inherits foo() from A
				// class C extends B {
				//   @Overide int foo() { print(2); }
				// }
			}
		}
	}

}